/*
 * search_bar.h
 * Copyright 2016 John Lindgren
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions, and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions, and the following disclaimer in the documentation
 *    provided with the distribution.
 *
 * This software is provided "as is" and without any warranty, express or
 * implied. In no event shall the authors be liable for any damages arising from
 * the use of this software.
 */

#ifndef SEARCH_BAR_H
#define SEARCH_BAR_H

#include <QWidget>

class PlaylistWidget;
class QLineEdit;

class SearchBar : public QWidget
{
public:
    SearchBar(QWidget * parent, PlaylistWidget * playlistWidget);

private:
    void keyPressEvent(QKeyEvent * event);

    PlaylistWidget * m_playlistWidget;
    QLineEdit * m_entry;
};

#endif
