/***************************************************************************
                            dma.h  -  description
                             -------------------
    begin                : Wed May 15 2002
    copyright            : (C) 2002 by Pete Bernert
    email                : BlackDove@addcom.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version. See also the license.txt file for *
 *   additional informations.                                              *
 *                                                                         *
 ***************************************************************************/

//*************************************************************************//
// History of changes:
//
// 2002/05/15 - Pete
// - generic cleanup for the Peops release
//
//*************************************************************************//

#include <stdint.h>

void InterruptDMA4(void);
void InterruptDMA7(void);

void SPU2readDMA4Mem(uint32_t usPSXMem,int iSize);
void SPU2writeDMA4Mem(uint32_t usPSXMem,int iSize);
void SPU2readDMA7Mem(uint32_t usPSXMem,int iSize);
void SPU2writeDMA7Mem(uint32_t usPSXMem,int iSize);
void SPU2interruptDMA4(void);
void SPU2interruptDMA7(void);
